if __name__ == "__main__":
    print(
        "Type annotations for boto3.{{ service_name.class_name }} {{ boto3_version }}\n"
        "Version:         {{ build_version }}\n"
        "Builder version: {{ builder_version }}\n"
        "Docs:            https://pypi.org/project/{{ service_name.pypi_name }}/\n"
        "Other services:  https://pypi.org/project/boto3-stubs/\n"
        "Changelog:       https://github.com/vemel/mypy_boto3_builder/releases"
    )
